/*
 * Copyright (C) 2010-2018 Apple Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include "NetworkProcessProxy.h"

#include "ArgumentCoders.h"
#include "Attachment.h"
#include "Decoder.h"
#include "HandleMessage.h"
#include "NetworkProcessProxyMessages.h"
#if ENABLE(SANDBOX_EXTENSIONS)
#include "SandboxExtension.h"
#endif
#if ENABLE(CONTENT_EXTENSIONS)
#include "UserContentControllerIdentifier.h"
#endif
#include "WebCoreArgumentCoders.h"
#include "WebsiteData.h"
#include <WebCore/AuthenticationChallenge.h>
#include <WebCore/DiagnosticLoggingClient.h>
#if ENABLE(SERVICE_WORKER)
#include <WebCore/SecurityOriginData.h>
#endif
#include <pal/SessionID.h>
#if ENABLE(RESOURCE_LOAD_STATISTICS) || ENABLE(SANDBOX_EXTENSIONS)
#include <wtf/Vector.h>
#endif
#include <wtf/text/WTFString.h>

namespace Messages {

namespace NetworkProcessProxy {

#if ENABLE(SANDBOX_EXTENSIONS)

void GetSandboxExtensionsForBlobFiles::callReply(IPC::Decoder& decoder, CompletionHandler<void(WebKit::SandboxExtension::HandleArray&&)>&& completionHandler)
{
    std::optional<WebKit::SandboxExtension::HandleArray> extensions;
    decoder >> extensions;
    if (!extensions) {
        ASSERT_NOT_REACHED();
        return;
    }
    completionHandler(WTFMove(*extensions));
}

void GetSandboxExtensionsForBlobFiles::cancelReply(CompletionHandler<void(WebKit::SandboxExtension::HandleArray&&)>&& completionHandler)
{
    completionHandler({ });
}

void GetSandboxExtensionsForBlobFiles::send(std::unique_ptr<IPC::Encoder>&& encoder, IPC::Connection& connection, const WebKit::SandboxExtension::HandleArray& extensions)
{
    *encoder << extensions;
    connection.sendSyncReply(WTFMove(encoder));
}

#endif

} // namespace NetworkProcessProxy

} // namespace Messages

namespace WebKit {

void NetworkProcessProxy::didReceiveNetworkProcessProxyMessage(IPC::Connection& connection, IPC::Decoder& decoder)
{
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidCreateNetworkConnectionToWebProcess::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidCreateNetworkConnectionToWebProcess>(decoder, this, &NetworkProcessProxy::didCreateNetworkConnectionToWebProcess);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidReceiveAuthenticationChallenge::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidReceiveAuthenticationChallenge>(decoder, this, &NetworkProcessProxy::didReceiveAuthenticationChallenge);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidFetchWebsiteData::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidFetchWebsiteData>(decoder, this, &NetworkProcessProxy::didFetchWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidDeleteWebsiteData::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidDeleteWebsiteData>(decoder, this, &NetworkProcessProxy::didDeleteWebsiteData);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidDeleteWebsiteDataForOrigins::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidDeleteWebsiteDataForOrigins>(decoder, this, &NetworkProcessProxy::didDeleteWebsiteDataForOrigins);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidSyncAllCookies::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidSyncAllCookies>(decoder, this, &NetworkProcessProxy::didSyncAllCookies);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidWriteBlobToFilePath::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidWriteBlobToFilePath>(decoder, this, &NetworkProcessProxy::didWriteBlobToFilePath);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::ProcessReadyToSuspend::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::ProcessReadyToSuspend>(decoder, this, &NetworkProcessProxy::processReadyToSuspend);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::SetIsHoldingLockedFiles::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::SetIsHoldingLockedFiles>(decoder, this, &NetworkProcessProxy::setIsHoldingLockedFiles);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::LogDiagnosticMessage::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::LogDiagnosticMessage>(decoder, this, &NetworkProcessProxy::logDiagnosticMessage);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::LogDiagnosticMessageWithResult::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::LogDiagnosticMessageWithResult>(decoder, this, &NetworkProcessProxy::logDiagnosticMessageWithResult);
        return;
    }
    if (decoder.messageName() == Messages::NetworkProcessProxy::LogDiagnosticMessageWithValue::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::LogDiagnosticMessageWithValue>(decoder, this, &NetworkProcessProxy::logDiagnosticMessageWithValue);
        return;
    }
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidUpdateBlockCookies::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidUpdateBlockCookies>(decoder, this, &NetworkProcessProxy::didUpdateBlockCookies);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidSetAgeCapForClientSideCookies::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidSetAgeCapForClientSideCookies>(decoder, this, &NetworkProcessProxy::didSetAgeCapForClientSideCookies);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::StorageAccessRequestResult::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::StorageAccessRequestResult>(decoder, this, &NetworkProcessProxy::storageAccessRequestResult);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::AllStorageAccessEntriesResult::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::AllStorageAccessEntriesResult>(decoder, this, &NetworkProcessProxy::allStorageAccessEntriesResult);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidRemoveAllStorageAccess::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidRemoveAllStorageAccess>(decoder, this, &NetworkProcessProxy::didRemoveAllStorageAccess);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidSetCacheMaxAgeCapForPrevalentResources::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidSetCacheMaxAgeCapForPrevalentResources>(decoder, this, &NetworkProcessProxy::didSetCacheMaxAgeCapForPrevalentResources);
        return;
    }
#endif
#if ENABLE(RESOURCE_LOAD_STATISTICS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::DidResetCacheMaxAgeCapForPrevalentResources::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::DidResetCacheMaxAgeCapForPrevalentResources>(decoder, this, &NetworkProcessProxy::didResetCacheMaxAgeCapForPrevalentResources);
        return;
    }
#endif
#if ENABLE(CONTENT_EXTENSIONS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::ContentExtensionRules::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::ContentExtensionRules>(decoder, this, &NetworkProcessProxy::contentExtensionRules);
        return;
    }
#endif
    if (decoder.messageName() == Messages::NetworkProcessProxy::RetrieveCacheStorageParameters::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::RetrieveCacheStorageParameters>(decoder, this, &NetworkProcessProxy::retrieveCacheStorageParameters);
        return;
    }
#if ENABLE(SANDBOX_EXTENSIONS)
    if (decoder.messageName() == Messages::NetworkProcessProxy::GetSandboxExtensionsForBlobFiles::name()) {
        IPC::handleMessageAsync<Messages::NetworkProcessProxy::GetSandboxExtensionsForBlobFiles>(connection, decoder, this, &NetworkProcessProxy::getSandboxExtensionsForBlobFiles);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcessProxy::EstablishWorkerContextConnectionToNetworkProcess::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::EstablishWorkerContextConnectionToNetworkProcess>(decoder, this, &NetworkProcessProxy::establishWorkerContextConnectionToNetworkProcess);
        return;
    }
#endif
#if ENABLE(SERVICE_WORKER)
    if (decoder.messageName() == Messages::NetworkProcessProxy::EstablishWorkerContextConnectionToNetworkProcessForExplicitSession::name()) {
        IPC::handleMessage<Messages::NetworkProcessProxy::EstablishWorkerContextConnectionToNetworkProcessForExplicitSession>(decoder, this, &NetworkProcessProxy::establishWorkerContextConnectionToNetworkProcessForExplicitSession);
        return;
    }
#endif
    UNUSED_PARAM(connection);
    UNUSED_PARAM(decoder);
    ASSERT_NOT_REACHED();
}

} // namespace WebKit

